#include "MotionControl.h"

// Core
#include <Core/robotstate/remote/RemoteRobot.h>
#include <Core/observers/variant/ChannelRef.h>
#include <Core/observers/ConditionCheck.h>
#include <Core/core/system/ArmarXDataPath.h>

// Virtual Robot
#include <VirtualRobot/IK/GazeIK.h>
#include <VirtualRobot/IK/GenericIKSolver.h>
#include <VirtualRobot/IK/DifferentialIK.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/MathTools.h>
#include <VirtualRobot/XML/RobotIO.h>

#include <Eigen/src/Geometry/Quaternion.h>

#include "ZeroForceControl.h"

using namespace armarx;
using namespace armarx::MotionControl;

MotionControlOfferer::MotionControlOfferer()
{
}



void MotionControlOfferer::onInitRemoteStateOfferer()
{
    addState<MoveJoints>("MoveJoints");
    //addState<MotionControlTestState>("MotionControlTestState");
    addState<MotionControlTestStateIK>("MotionControlTestStateIK");
    addState<ZeroForceControl>("ZeroForceControl");
}



void MotionControlOfferer::onConnectRemoteStateOfferer()
{
}



void MoveJoints::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    //setConfigFile("RobotAPI/data/stateconfigs/MotionControl.xml");

    addToInput("jointNames", VariantType::List(VariantType::String), false);
    addToInput("targetJointValues", VariantType::List(VariantType::Float), false);

    // TODO: add when we have decided how to do it
    addToInput("jointMaxSpeeds", VariantType::List(VariantType::Float), true);
    //addToInput("jointMaxSpeed", VariantType::List(VariantType::Float), true);
    addToInput("jointMaxSpeed", VariantType::Float, true);

    addToInput("jointTargetTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);

    addToOutput("jointDistancesToTargets", VariantType::List(VariantType::Float), true);
}



void MoveJoints::onEnter()
{
    NameValueMap targetJointAngles;
    SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
    SingleTypeVariantListPtr targetJointValues = getInput<SingleTypeVariantList>("targetJointValues");

    if (jointNames->getSize()!=targetJointValues->getSize())
    {
        throw LocalException("Sizes of joint name list and joint value list do not match!");
    }

    ARMARX_LOG << "number of joints that will be set: " << jointNames->getSize() << flush;

    for (int i=0; i<jointNames->getSize(); i++)
    {
        targetJointAngles[jointNames->getVariant(i)->getString()] = targetJointValues->getVariant(i)->getFloat();
        ARMARX_VERBOSE << "setting joint angle for joint '" << jointNames->getVariant(i)->getString() << "' to " << targetJointValues->getVariant(i)->getFloat() << std::endl;
    }
    // TODO: Set Max Velocities

    // now install the condition
    Term cond;
    float tolerance = getInput<float>("jointTargetTolerance");
    for (int i=0; i<jointNames->getSize(); i++)
    {
        ARMARX_VERBOSE << "creating condition (" << i << " of " << jointNames->getSize() << ")" << flush;
        ParameterList inrangeCheck;
        inrangeCheck.push_back(new Variant(targetJointValues->getVariant(i)->getFloat()-tolerance));
        inrangeCheck.push_back(new Variant(targetJointValues->getVariant(i)->getFloat()+tolerance));

        Literal inrangeLiteral(DataFieldIdentifier(context->getKinematicUnitObserverName(), "jointangles", jointNames->getVariant(i)->getString()), "inrange", inrangeCheck);
        cond = cond && inrangeLiteral;
    }

    ARMARX_INFO << "installing condition: EvJointTargetReached" << std::endl;
    targetReachedCondition = installCondition<EvJointTargetReached>(cond);
    ARMARX_INFO << "condition installed: EvJointTargetReached" << std::endl;
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    ARMARX_INFO << "timeout installed" << std::endl;
    context->kinematicUnitPrx->setJointAngles(targetJointAngles);
}



void MoveJoints::onBreak()
{
    removeCondition(targetReachedCondition);
    removeTimeoutEvent(timeoutEvent);
}



void MoveJoints::onExit()
{
    SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
    SingleTypeVariantListPtr targetJointValues = getInput<SingleTypeVariantList>("targetJointValues");
    SingleTypeVariantList resultList(VariantType::Float);

    for(int i=0; i<jointNames->getSize(); i++ )
    {
        DataFieldIdentifierPtr datafield = new DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointangles", jointNames->getVariant(i)->getString());
        float jointAngleActual = context->kinematicUnitObserverPrx->getDataField(datafield)->getFloat();
        float jointTargetValue = targetJointValues->getVariant(i)->getFloat();
        resultList.addVariant(Variant(jointAngleActual - jointTargetValue));
    }
    setOutput("jointDistancesToTargets", resultList);

    ARMARX_INFO << "output: " << StateUtilFunctions::getDictionaryString(getOutputParameters()) << std::endl;

    removeCondition(targetReachedCondition);
    removeTimeoutEvent(timeoutEvent);
}



void MoveJointsVelocityControl::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    //setConfigFile("RobotAPI/data/stateconfigs/MotionControl.xml");

    addToInput("jointNames", VariantType::List(VariantType::String), false);
    addToInput("targetJointVelocities", VariantType::List(VariantType::Float), false);

    // TODO: add when we have decided how to do it
    addToInput("jointMaxSpeeds", VariantType::List(VariantType::Float), true);
    addToInput("jointMaxSpeed", VariantType::Float, true);

    addToInput("accelerationTime", VariantType::Float, false);

    addToInput("targetJointVelocityTolerance", VariantType::Float, false);

    addToInput("timeoutInMs", VariantType::Int, false);

    addToOutput("jointVelocitiesDistancesToTargets", VariantType::Float, true);
}



void MoveJointsVelocityControl::onEnter()
{
    //float accelerationTime = getInput<float>("accelerationTime");
    NameValueMap targetJointVelocities;
    SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
    SingleTypeVariantListPtr targetVelocitiesValues = getInput<SingleTypeVariantList>("targetJointVelocities");

    if (jointNames->getSize()!=targetVelocitiesValues->getSize())
    {
        throw LocalException("Sizes of joint name list and joint velocities list do not match!");
    }

    for(int i=0; i<jointNames->getSize(); i++)
    {
        targetJointVelocities[jointNames->getVariant(i)->getString()] = targetVelocitiesValues->getVariant(i)->getFloat();
        ARMARX_VERBOSE << "setting joint velocity for joint '" << jointNames->getVariant(i)->getString() << "' to " << targetVelocitiesValues->getVariant(i)->getFloat() << std::endl;
    }
    // TODO: Set Max Velocities

    // now install the condition

    Term cond;
    float tolerance = getInput<float>("targetJointVelocityTolerance");
    for(int i=0; i<jointNames->getSize(); i++)
    {
        ParameterList inrangeCheck;
        inrangeCheck.push_back(new Variant(targetVelocitiesValues->getVariant(i)->getFloat()-tolerance));
        inrangeCheck.push_back(new Variant(targetVelocitiesValues->getVariant(i)->getFloat()+tolerance));

        Literal inrangeLiteral(DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointvelocities", jointNames->getVariant(i)->getString()), "inrange", inrangeCheck);
        cond = cond && inrangeLiteral;
    }

    targetVelocitiesReachedCondition = installCondition<EvJointVelocitiesTargetReached>(cond);
    ARMARX_INFO << "condition installed" << std::endl;
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    ARMARX_INFO << "timeout installed" << std::endl;
//    context->kinematicUnitPrx->setAccelerationTime(accelerationTime);
    context->kinematicUnitPrx->setJointVelocities(targetJointVelocities);
}



void MoveJointsVelocityControl::onExit()
{
    SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
    SingleTypeVariantListPtr targetJointVelocities = getInput<SingleTypeVariantList>("targetJointVelocities");
    SingleTypeVariantList resultList(VariantType::Float);

    for(int i=0; i<jointNames->getSize(); i++)
    {
        DataFieldIdentifierPtr datafield = new DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointvelocities", jointNames->getVariant(i)->getString());
        float jointVelocityActual = context->kinematicUnitObserverPrx->getDataField(datafield)->getFloat();
        float jointVelocityTarget = targetJointVelocities->getVariant(i)->getFloat();
        resultList.addVariant(Variant(jointVelocityActual - jointVelocityTarget));
    }
    setOutput("jointVelocitiesDistancesToTargets", resultList);

    ARMARX_INFO << "output: " << StateUtilFunctions::getDictionaryString(getOutputParameters()) << std::endl;

    removeCondition(targetVelocitiesReachedCondition);
    removeTimeoutEvent(timeoutEvent);
}



void MoveTCPPoseIK::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    //setConfigFile("RobotAPI/data/stateconfigs/MotionControl.xml");

    addToInput("kinematicChainName", VariantType::String, false);
    addToInput("targetTCPPosition", VariantType::FramedPosition, false);
    addToInput("targetTCPOrientation", VariantType::FramedOrientation, false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);

    addToInput("jointTargetTolerance", VariantType::Float, false);

    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void MoveTCPPoseIK::onEnter()
{
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
}



void MoveTCPPoseIK::onExit()
{
    removeTimeoutEvent(timeoutEvent);
}



void MoveTCPPoseIK::defineSubstates()
{
    StatePtr calculateJointAngleConfiguration = addState<CalculateJointAngleConfiguration>("CalculateJointAngleConfiguration");
    StatePtr moveJoints = addState<MoveJoints>("MoveJoints");
    StatePtr validateJointAngleConfiguration = addState<ValidateJointAngleConfiguration>("ValidateJointAngleConfiguration");
    StatePtr movingDone = addState<SuccessState>("movingDone") ;
    StatePtr movingFailed = addState<FailureState>("movingFailed") ;

    setInitState(calculateJointAngleConfiguration, createMapping()->mapFromParent("*","*")); // PM::createMapping()->mapFromParent("*","*") ??
    addTransition<EvCalculationDone>(calculateJointAngleConfiguration, moveJoints, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvJointTargetReached>(moveJoints, validateJointAngleConfiguration, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvSuccess>(validateJointAngleConfiguration, movingDone, createMapping()->mapFromOutput("*","*"));
    addTransitionFromAllStates<EvTimeout>(movingFailed);
    addTransitionFromAllStates<EvFailure>(movingFailed);
}



void MoveTCPTrajectory::defineParameters()
{
    context = getContext<RobotStatechartContext>();

    //setConfigFile("RobotAPI/data/stateconfigs/MotionControl.xml");

    addToInput("kinematicChainName", VariantType::String, false);
    addToInput("targetTCPPositions", VariantType::List(VariantType::FramedPosition), false);
    addToInput("targetTCPOrientations", VariantType::List(VariantType::FramedOrientation), false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);


    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);

    addToLocal("trajectoryPointCounterID", VariantType::ChannelRef);
}



void MoveTCPTrajectory::onEnter()
{
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    int numberOfPoints = getInput<SingleTypeVariantList>("targetTCPPositions")->getSize();
    if (numberOfPoints == 0)
    {
        sendEvent<EvSuccess>();
    }
    else if (numberOfPoints != getInput<SingleTypeVariantList>("targetTCPOrientations")->getSize())
    {
        throw armarx::exceptions::local::eStatechartLogicError("the number of targetTCPPositions does not equal the number of targetTCPOrientations");
        sendEvent<EvFailure>();
    }
}



void MoveTCPTrajectory::onExit()
{
    removeTimeoutEvent(timeoutEvent);
    ChannelRefPtr r = getLocal<ChannelRef>("trajectoryPointCounterID");
    getContext()->systemObserverPrx->removeCounter(r);
    //removeCounterEvent(trajectoryPointCounterID);
}



void MoveTCPTrajectory::defineSubstates()
{
    StatePtr moveTCPTrajectoryInit = addState<MoveTCPTrajectoryInit>("MoveTCPTrajectoryInit");
    StatePtr moveTCPTrajectoryCheckCounter = addState<MoveTCPTrajectoryCheckCounter>("MoveTCPTrajectoryCheckCounter");
    StatePtr moveTCPTrajectoryNextPoint = addState<MoveTCPTrajectoryNextPoint>("MoveTCPTrajectoryNextPoint");
    StatePtr movingDone = addState<SuccessState>("movingDone");
    StatePtr movingFailed = addState<FailureState>("movingFailed");
    setInitState(moveTCPTrajectoryInit, PM::createMapping()->mapFromParent("*","*"));
    addTransition<EvSuccess>(moveTCPTrajectoryInit, moveTCPTrajectoryCheckCounter, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvLastPointNotYetReached>(moveTCPTrajectoryCheckCounter, moveTCPTrajectoryNextPoint, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvSuccess>(moveTCPTrajectoryNextPoint, moveTCPTrajectoryCheckCounter, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransition<EvLastPointReached>(moveTCPTrajectoryCheckCounter, movingDone, createMapping()->mapFromParent("*","*")->mapFromOutput("*","*"));
    addTransitionFromAllStates<EvTimeout>(movingFailed);
    addTransitionFromAllStates<EvFailure>(movingFailed);
}




void MoveTCPTrajectoryInit::defineParameters()
{
    addToInput("targetTCPPositions", VariantType::List(VariantType::FramedPosition), false);
    addToOutput("trajectoryPointCounterID", VariantType::ChannelRef, false);
}




void MoveTCPTrajectoryInit::onEnter()
{
    // create a counter to keep track of the poses that were already reached. As we dont use the event, set the max to n+1 so it will never be sent
    int numberOfPoints = getInput<SingleTypeVariantList>("targetTCPPositions")->getSize();
    ARMARX_INFO << "number of points on trajectory: " << numberOfPoints;
    ActionEventIdentifier trajectoryPointCounterID = setCounterEvent(numberOfPoints, createEvent<EvLastPointReached>());
    setOutput("trajectoryPointCounterID", trajectoryPointCounterID.actionId.get());

    sendEvent<EvSuccess>();
}




void MoveTCPTrajectoryCheckCounter::defineParameters()
{
    addToInput("targetTCPPositions", VariantType::List(VariantType::FramedPosition), false);
    addToInput("trajectoryPointCounterID", VariantType::ChannelRef, false);
    addToOutput("trajectoryPointCounterID", VariantType::ChannelRef, false);
}




void MoveTCPTrajectoryCheckCounter::onEnter()
{
    context = getContext<RobotStatechartContext>();
    ChannelRefPtr trajectoryPointCounterID = getInput<ChannelRef>("trajectoryPointCounterID");
    setOutput("trajectoryPointCounterID", trajectoryPointCounterID);

    int counterValue = trajectoryPointCounterID->getDataField("value")->getInt();

    ARMARX_INFO << "counter: " << counterValue;

    if (counterValue < getInput<SingleTypeVariantList>("targetTCPPositions")->getSize())
    {
        sendEvent<EvLastPointNotYetReached>();
    }
    else
    {
        sendEvent<EvLastPointReached>();
    }
}





void MoveTCPTrajectoryNextPoint::defineParameters()
{
    addToInput("kinematicChainName", VariantType::String, false);
    addToInput("targetTCPPositions", VariantType::List(VariantType::FramedPosition), false);
    addToInput("targetTCPOrientations", VariantType::List(VariantType::FramedOrientation), false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);

    addToInput("trajectoryPointCounterID", VariantType::ChannelRef, false);
    addToOutput("trajectoryPointCounterID", VariantType::ChannelRef, false);


    addToLocal("targetTCPPosition", VariantType::FramedPosition);
    addToLocal("targetTCPOrientation", VariantType::FramedOrientation);


    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void MoveTCPTrajectoryNextPoint::onEnter()
{
    context = getContext<RobotStatechartContext>();
    ChannelRefPtr trajectoryPointCounterID = getInput<ChannelRef>("trajectoryPointCounterID");
    setOutput("trajectoryPointCounterID", trajectoryPointCounterID);

    int counterValue = trajectoryPointCounterID->getDataField("value")->getInt();

    setLocal("targetTCPPosition", *VariantPtr::dynamicCast(getInput<SingleTypeVariantList>("targetTCPPositions")->getVariant(counterValue)));
    setLocal("targetTCPOrientation", *VariantPtr::dynamicCast(getInput<SingleTypeVariantList>("targetTCPOrientations")->getVariant(counterValue)));

    context->systemObserverPrx->incrementCounter(trajectoryPointCounterID);
}




void MoveTCPTrajectoryNextPoint::defineSubstates()
{
    StatePtr moveTCPPoseIK = addState<MoveTCPPoseIK>("MoveTCPPoseIK");
    StatePtr movingDone = addState<SuccessState>("movingDone") ;
    StatePtr movingFailed = addState<FailureState>("movingFailed") ;
    setInitState(moveTCPPoseIK, PM::createMapping()->mapFromParent("*","*"));
    addTransition<EvSuccess>(moveTCPPoseIK, movingDone);
    addTransitionFromAllStates<EvTimeout>(movingFailed);
    addTransitionFromAllStates<EvFailure>(movingFailed);
}



void CalculateJointAngleConfiguration::defineParameters()
{
    addToInput("kinematicChainName", VariantType::String, false);
    addToInput("targetTCPPosition", VariantType::FramedPosition, false);
    addToInput("targetTCPOrientation", VariantType::FramedOrientation, false);
    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);

    addToOutput("jointNames", VariantType::List(VariantType::String), true);
    addToOutput("targetJointValues", VariantType::List(VariantType::Float), true);
}



void CalculateJointAngleConfiguration::run()
{
    //RobotStatechartContext* context = getContext<RobotStatechartContext>();
    //VirtualRobot::RobotPtr robot(new RemoteRobot(context->robotStateComponent->getRobotSnapshot("CalculateTCPPoseTime")));

    // TODO: with the following line, the IK doesn't find a solution, so this terrible hack must be used. Fix it!!!
    //VirtualRobot::RobotPtr robot = robotPtr->clone("CalculateTCPPoseClone");
    std::string robotModelFile;
    //ArmarXDataPath::getAbsolutePath("Armar4/data/Armar4/ArmarIV.xml", robotModelFile);
    ArmarXDataPath::getAbsolutePath("Armar3/data/robotmodel/ArmarIII.xml", robotModelFile);
    VirtualRobot::RobotPtr robot = VirtualRobot::RobotIO::loadRobot(robotModelFile.c_str());

    VirtualRobot::GenericIKSolverPtr ikSolver(new VirtualRobot::GenericIKSolver(robot->getRobotNodeSet(getInput<std::string>("kinematicChainName"))));
    ikSolver->setVerbose(true);
    ikSolver->setMaximumError(getInput<float>("targetPositionDistanceTolerance"));
    ikSolver->setupJacobian(0.1, 40);

    VirtualRobot::LinkedCoordinate target = ArmarPose::createLinkedCoordinate(robot, getInput<FramedPosition>("targetTCPPosition"), getInput<FramedOrientation>("targetTCPOrientation"));
    ARMARX_INFO << "target: " << target.getPose() << armarx::flush;
    Eigen::Matrix4f goalInRoot = target.getInFrame(robot->getRootNode());

//    // test
//    VirtualRobot::RobotNodePtr rn = robot->getRobotNode("LeftTCP");
//    Eigen::Matrix4f goalTmpTCP = rn->getGlobalPose();
    Eigen::Matrix4f goalGlobal = robot->getRootNode()->toGlobalCoordinateSystem(goalInRoot);
    ARMARX_INFO << "GOAL in root: " << goalInRoot << armarx::flush;
    ARMARX_INFO << "GOAL global: " << goalGlobal << armarx::flush;
//    ARMARX_INFO << "GOAL TCP:" << goalTmpTCP << endl;

    ARMARX_INFO << "Calculating IK" << flush;
    //if (!ikSolver->solve(goalGlobal, VirtualRobot::IKSolver::Position, 50))
    if (!ikSolver->solve(goalGlobal, VirtualRobot::IKSolver::All, 50))
    {
        ARMARX_WARNING << "IKSolver found no solution! " << flush;
        sendEvent<EvFailure>();
    }
    else
    {

        SingleTypeVariantList jointNames = SingleTypeVariantList(VariantType::String);
        SingleTypeVariantList targetJointValues = SingleTypeVariantList(VariantType::Float);
        VirtualRobot::RobotNodeSetPtr kinematicChain = robot->getRobotNodeSet(getInput<std::string>("kinematicChainName") );
        for (int i = 0; i < (signed int)kinematicChain->getSize(); i++)
        {
            if (kinematicChain->getNode(i)->getName().compare("VirtualCentralGaze") != 0)
            {
                jointNames.addVariant(Variant(kinematicChain->getNode(i)->getName()));
                targetJointValues.addVariant(Variant(kinematicChain->getNode(i)->getJointValue()));
                ARMARX_LOG << " joint: " << kinematicChain->getNode(i)->getName() << "   value: " << kinematicChain->getNode(i)->getJointValue() << flush;
            }
        }
        ARMARX_LOG << "number of joints to be set: " << jointNames.getSize() << flush;
        ARMARX_LOG << "setting output: jointNames" << flush;
        setOutput("jointNames", jointNames);
        ARMARX_LOG << "setting output: targetJointValues" << flush;
        setOutput("targetJointValues", targetJointValues);
        sendEvent<EvCalculationDone>();
    }
}



void ValidateJointAngleConfiguration::defineParameters()
{
    addToInput("targetTCPPosition", VariantType::FramedPosition, false);
    addToInput("targetTCPOrientation", VariantType::FramedOrientation, false);
    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("kinematicChainName", VariantType::String, false);

    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void ValidateJointAngleConfiguration::onEnter()
{
    using namespace Eigen;

    RobotStatechartContext* context = getContext<RobotStatechartContext>();
    VirtualRobot::RobotPtr robot(new RemoteRobot(context->robotStateComponent->getRobotSnapshot("ValidateTCPPoseTime")));
    VirtualRobot::LinkedCoordinate actualPose(robot);
    actualPose.set(robot->getRobotNodeSet(getInput<std::string>("kinematicChainName"))->getTCP()->getName(), Vector3f(0,0,0));
    actualPose.changeFrame(robot->getRootNode());
    Vector3f actualPosition = actualPose.getPosition();
    Quaternionf actualOrientation(actualPose.getPose().block<3,3>(0,0));


    VirtualRobot::LinkedCoordinate targetPose = ArmarPose::createLinkedCoordinate(robot, getInput<FramedPosition>("targetTCPPosition"), getInput<FramedOrientation>("targetTCPOrientation"));
    targetPose.changeFrame(robot->getRootNode());
    Vector3f targetPosition = targetPose.getPosition();
    Quaternionf targetOrientation(targetPose.getPose().block<3,3>(0,0));

    float cartesianDistance = sqrtf(((targetPosition-actualPosition).dot(targetPosition-actualPosition)));
    float orientationDistance = actualOrientation.angularDistance(targetOrientation);

    setOutput("TCPDistanceToTarget", cartesianDistance);
    setOutput("TCPOrientationDistanceToTarget", orientationDistance);
    if (cartesianDistance <= getInput<float>("targetPositionDistanceTolerance") && orientationDistance <= getInput<float>("targetOrientationDistanceTolerance"))
    {
        sendEvent<EvSuccess>();
    }
    else
    {
        ARMARX_WARNING << "Target pose has not been reached with the desired accuracy. Cartesian distance: " << cartesianDistance << ", orientation distance: " << orientationDistance << " ";
        ARMARX_INFO << "Actual pose: " << actualPose.getPose();
        ARMARX_INFO << "Target pose: " << targetPose.getPose();
        sendEvent<EvSuccess>();
    }
}



void MotionControlTestState::defineParameters()
{
    //setConfigFile("RobotAPIConfigs/stateconfigs/MotionControl.xml");

    addToInput("jointNames", VariantType::List(VariantType::String), false);
    addToInput("targetJointValues", VariantType::List(VariantType::Float), false);

    addToInput("jointMaxSpeed", VariantType::Float, true);

    addToInput("jointTargetTolerance", VariantType::Float, false);

    addToInput("timeoutInMs", VariantType::Int, false);

    addToOutput("jointDistancesToTargets", VariantType::Float, true);
}



void MotionControlTestState::defineSubstates()
{
    StatePtr moveJoints = addState<MoveJoints>("MoveJoints");
    StatePtr movingDone = addState<SuccessState>("movingDone");
    StatePtr movingFailed = addState<FailureState>("movingFailed");

    setInitState(moveJoints, PM::createMapping()->mapFromParent("*","*"));

    addTransition<EvJointTargetReached>(moveJoints, movingDone);
    addTransition<EvTimeout>(moveJoints, movingFailed);
}



void MotionControlTestStateIK::defineParameters()
{
    //setConfigFile("RobotAPIConfigs/stateconfigs/MotionControl.xml");

    addToInput("kinematicChainName", VariantType::String, false);
    //addToInput("targetTCPPosition", VariantType::FramedPosition, false);
    //addToInput("targetTCPOrientation", VariantType::FramedOrientation, false);
    addToInput("targetTCPPositions", VariantType::List(VariantType::FramedPosition), false);
    addToInput("targetTCPOrientations", VariantType::List(VariantType::FramedOrientation), false);

    addToInput("tcpMaxSpeed", VariantType::Float, true);

    addToInput("targetPositionDistanceTolerance", VariantType::Float, false);
    addToInput("targetOrientationDistanceTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);


    addToOutput("TCPDistanceToTarget", VariantType::Float, true);
    addToOutput("TCPOrientationDistanceToTarget", VariantType::Float, true);
}



void MotionControlTestStateIK::defineSubstates()
{
    //StatePtr moveTCPPoseIK = addState<MoveTCPPoseIK>("MoveTCPPoseIK");
    StatePtr moveTCPTrajectory = addState<MoveTCPTrajectory>("MoveTCPTrajectory");
    StatePtr movingDone = addState<SuccessState>("movingDone");
    StatePtr movingFailed = addState<FailureState>("movingFailed");

    //setInitState(moveTCPPoseIK, PM::createMapping()->mapFromParent("*","*"));
    setInitState(moveTCPTrajectory, PM::createMapping()->mapFromParent("*","*"));

    //addTransition<EvSuccess>(moveTCPPoseIK, movingDone, PM::createMapping()->mapFromOutput("*","*"));
    //addTransition<EvTimeout>(moveTCPPoseIK, movingFailed);
    //addTransition<EvFailure>(moveTCPPoseIK, movingFailed);
    addTransition<EvSuccess>(moveTCPTrajectory, movingDone, PM::createMapping()->mapFromOutput("*","*"));
    addTransition<EvTimeout>(moveTCPTrajectory, movingFailed);
    addTransition<EvFailure>(moveTCPTrajectory, movingFailed);
}



void MotionControl::StopRobot::onEnter()
{
    ARMARX_LOG << "entering MotionControl::StopRobot" << flush;

    RobotStatechartContext* context = getContext<RobotStatechartContext>();

    //VirtualRobot::RobotPtr robot = (RemoteRobot(context->robotStateComponent->getRobotSnapshot("StopRobotTime"))).clone("StopRobotCopy"); this crashes
    //std::vector<VirtualRobot::RobotNodePtr> allNodes;
    //robot->getRobotNodes(allNodes);

    armarx::NameList allNodes = context->robotStateComponent->getRobotSnapshot("StopRobotTime")->getRobotNodes();

    NameValueMap jointVelocities;
    //std::vector<VirtualRobot::RobotNodePtr>::iterator it = allNodes.begin();
    armarx::NameList::iterator it = allNodes.begin();
    for(int i=0; it != allNodes.end(); it++, i++)
    {
        //jointVelocities[(*it)->getName()] = 0.0f;
        jointVelocities[*it] = 0.0f;
    }

    context->kinematicUnitPrx->begin_setJointVelocities(jointVelocities);

    sendEvent<EvSuccess>();
}




void CloseHand::defineParameters()
{
    addToInput("useLeftHand", VariantType::Bool, false);
}


void CloseHand::onEnter()
{
    RobotStatechartContext* context = getContext<RobotStatechartContext>();
    bool useLeftHand = getInput<bool>("useLeftHand");
    NameValueMap handConfig;
    if (useLeftHand)
    {
        handConfig["left_hand_configuration_actual_float"] = 4;
    }
    else
    {
        handConfig["right_hand_configuration_actual_float"] = 4;
    }

    context->kinematicUnitPrx->setJointAngles(handConfig);

    sendEvent<EvSuccess>();
}




void OpenHand::defineParameters()
{
    addToInput("useLeftHand", VariantType::Bool, false);
}


void OpenHand::onEnter()
{
    RobotStatechartContext* context = getContext<RobotStatechartContext>();
    bool useLeftHand = getInput<bool>("useLeftHand");
    NameValueMap handConfig;
    if (useLeftHand)
    {
        handConfig["left_hand_configuration_actual_float"] = 1;
    }
    else
    {
        handConfig["right_hand_configuration_actual_float"] = 1;
    }

    context->kinematicUnitPrx->setJointAngles(handConfig);

    sendEvent<EvSuccess>();
}





void HeadLookAtTarget::defineParameters()
{
    addToInput("target", VariantType::FramedPosition, false);
    addToInput("headIKKinematicChainName", VariantType::String, false);
    addToInput("jointTargetTolerance", VariantType::Float, false);
    addToInput("timeoutInMs", VariantType::Int, false);
}


void HeadLookAtTarget::defineSubstates()
{
    StatePtr stateCalculateHeadIK = addState<CalculateHeadIK>("CalculateHeadIKState");
    StatePtr stateMoveHeadJoints = addState<MoveJoints>("MoveHeadJointsState");
    StatePtr stateSuccess = addState<SuccessState>("Success");
    StatePtr stateFailure = addState<FailureState>("Failure");

    ParameterMappingPtr initialMapping = ParameterMapping::createMapping()->mapFromParent("*", "*");
    ParameterMappingPtr transitionMapping = ParameterMapping::createMapping()->mapFromOutput("*", "*")->mapFromParent("*", "*");

    setInitState(stateCalculateHeadIK, initialMapping);
    addTransition<EvCalculationDone>(stateCalculateHeadIK, stateMoveHeadJoints, transitionMapping);
    addTransition<EvJointTargetReached>(stateMoveHeadJoints, stateSuccess, transitionMapping);
    addTransitionFromAllStates<EvFailure>(stateFailure);
    addTransitionFromAllStates<EvTimeout>(stateFailure);
}




void CalculateHeadIK::defineParameters()
{
    addToInput("target", VariantType::FramedPosition, false);
    addToInput("headIKKinematicChainName", VariantType::String, false);

    addToOutput("jointNames", VariantType::List(VariantType::String), true);
    addToOutput("targetJointValues", VariantType::List(VariantType::Float), true);
}


void CalculateHeadIK::onEnter()
{
    targetPosition = getInput<FramedPosition>("target");
    kinematicChainName = getInput<std::string>("headIKKinematicChainName");
}


void CalculateHeadIK::run()
{
    RobotStatechartContext* context = getContext<RobotStatechartContext>();
    VirtualRobot::RobotPtr robotSnapshot(new RemoteRobot(context->robotStateComponent->getRobotSnapshot("CalculateTCPPoseTime")));
    // TODO: with the following line, the IK doesn't find a solution, so this terrible hack must be used. Fix it!!!
    //VirtualRobot::RobotPtr robot = robotPtr->clone("CalculateTCPPoseClone");
    VirtualRobot::RobotConfigPtr robotSnapshotConfig = robotSnapshot->getConfig();
    std::string robotModelFile;
    ArmarXDataPath::getAbsolutePath("Armar3/data/robotmodel/ArmarIII.xml", robotModelFile);
    VirtualRobot::RobotPtr robot = VirtualRobot::RobotIO::loadRobot(robotModelFile.c_str(), VirtualRobot::RobotIO::eStructure);
    robot->setConfig(robotSnapshotConfig);

    VirtualRobot::RobotNodeSetPtr kinematicChain = robot->getRobotNodeSet(kinematicChainName);
    VirtualRobot::RobotNodePrismaticPtr virtualJoint;
    for (unsigned int i=0; i<kinematicChain->getSize(); i++)
    {
        if (kinematicChain->getNode(i)->getName().compare("VirtualCentralGaze") == 0)
        {
            virtualJoint =  boost::dynamic_pointer_cast<VirtualRobot::RobotNodePrismatic>(kinematicChain->getNode(i));
        }
    }
    VirtualRobot::GazeIK ikSolver(kinematicChain, virtualJoint);
    ikSolver.enableJointLimitAvoidance(true);

    Eigen::Matrix3f m = Eigen::Matrix3f::Identity();
    FramedOrientationPtr targetOri = new FramedOrientation(m, targetPosition->frame);
    VirtualRobot::LinkedCoordinate target = ArmarPose::createLinkedCoordinate(robot, targetPosition, targetOri);
    ARMARX_INFO << "Target: " << target.getPose() << armarx::flush;
    Eigen::Matrix4f goalInRoot = target.getInFrame(robot->getRootNode());
    //Eigen::Matrix4f goalGlobal = robot->getRootNode()->toGlobalCoordinateSystem(goalInRoot);
    Eigen::Vector3f targetPos = goalInRoot.block<3,1>(0,3);

    ARMARX_INFO << "Calculating IK for target position " << targetPos;
    if (!ikSolver.solve(targetPos))
    {
        ARMARX_WARNING << "IKSolver found no solution! ";
        sendEvent<EvFailure>();
    }
    else
    {

        SingleTypeVariantList jointNames = SingleTypeVariantList(VariantType::String);
        SingleTypeVariantList targetJointValues = SingleTypeVariantList(VariantType::Float);
        for (int i = 0; i < (signed int)kinematicChain->getSize(); i++)
        {
            if (kinematicChain->getNode(i)->getName().compare("VirtualCentralGaze") != 0)
            {
                jointNames.addVariant(Variant(kinematicChain->getNode(i)->getName()));
                targetJointValues.addVariant(Variant(kinematicChain->getNode(i)->getJointValue()));
                ARMARX_LOG << " joint: " << kinematicChain->getNode(i)->getName() << "   value: " << kinematicChain->getNode(i)->getJointValue() << flush;
            }
        }
        ARMARX_LOG << "number of joints to be set: " << jointNames.getSize() << flush;
        ARMARX_LOG << "setting output: jointNames" << flush;
        setOutput("jointNames", jointNames);
        ARMARX_LOG << "setting output: targetJointValues" << flush;
        setOutput("targetJointValues", targetJointValues);
        sendEvent<EvCalculationDone>();
    }
}


